#include <QDebug>
#include <QList>
#include <assert.h>
#include "PluginList.h"
#include "Globals.h"

PluginList* PluginList::single        = NULL;
bool PluginList::       disableAlways = false;

int
PluginList::getCubePluginCount()
{
    if ( !single )
    {
        single = new PluginList();
    }
    return single->cubePluginList.size();
}

CubePlugin*
PluginList::getCubePlugin( int idx )
{
    assert( single );
    return static_cast<CubePlugin*> ( single->cubePluginList.at( idx ).plugin );
}

QString
PluginList::getCubePluginName( int idx )
{
    assert( single );
    return single->cubePluginList.at( idx ).name;
}

int
PluginList::getContextFreePluginCount()
{
    if ( !single )
    {
        single = new PluginList();
    }
    return single->contextFreePluginList.size();
}

ContextFreePlugin*
PluginList::getContextFreePlugin( int idx )
{
    assert( single );
    return static_cast<ContextFreePlugin*> ( single->contextFreePluginList.at( idx ).plugin );
}

QString
PluginList::getContextFreePluginName( int idx )
{
    assert( single );
    return single->contextFreePluginList.at( idx ).name;
}

// ------------

PluginList::PluginList()
{
    loadPluginList();
}

int
getVersionNumber( PluginInterface* plugin )
{
    int major, minor, bugfix;
    plugin->version( major, minor, bugfix );
    return major * 10000 + minor * 100 + bugfix;
}

/**
 * @brief PluginList::loadPluginList initial load of all available plugins
 * @param idx
 */
void
PluginList::loadPluginList()
{
    if ( disableAlways )
    {
        return;
    }

    QList<QDir> pluginDirList;

    // add user defined cube plugin directory, if environment variable "CUBE_PLUGIN_DIR" is set
    if ( qgetenv( "CUBE_PLUGIN_DIR" ).size() > 0 )
    {
        pluginDirList.append( QDir( QString( qgetenv( "CUBE_PLUGIN_DIR" ) ) ) );
    }
    else
    {
        Globals::debug( "PluginList" ) << "no user defined plugin path CUBE_PLUGIN_DIR" << endl;
    }
    pluginDirList.append( qApp->applicationDirPath() );
    pluginDirList.append( QDir( QString( LIBDIR ).append( "/plugins" ) ) );
    pluginDirList.append( QDir( QString( LIBDIR ).append( "/cube-plugins" ) ) );

    Globals::debug( "PluginList" ) << "plugin search path: " << endl;
    foreach( QDir dir, pluginDirList )
    {
        Globals::debug( "PluginList" ) << " " << dir.absolutePath() << endl;
    }

    foreach( QDir dir, pluginDirList )
    {
        QStringList filters;
        filters << "*.so" << "*.dll";
        dir.setNameFilters( filters );
        foreach( QString fileName, dir.entryList( QDir::Files ) )
        {
            QString        location = dir.absoluteFilePath( fileName );
            QPluginLoader* loader   = new QPluginLoader( location );
            QObject*       plugin   = loader->instance();

            if ( plugin ) // plugin has been loaded
            {
                PluginInterface* validPlugin = qobject_cast<PluginInterface*>( plugin );
                if ( validPlugin ) // plugin implements PluginInterface
                {
                    PluginData data;
                    data.location     = location;
                    data.plugin       = validPlugin;
                    data.pluginLoader = loader;
                    data.name         = data.plugin->name();

                    if ( loadContextFreePlugin( data ) )
                    {
                        Globals::debug( "PluginList" ) << "context free plugin loaded:" << data.location << endl;
                    }
                    else if ( loadCubePlugin( data ) )
                    {
                        Globals::debug( "PluginList" ) << "plugin loaded:" << data.location << endl;
                    }
                    else
                    {
                        validPlugin = 0;
                    }
                }
                if ( !validPlugin ) // unload invalid plugin
                {
                    loader->unload();
                    delete loader;
                    Globals::debug( "PluginList" ) << "plugin " << dir.absoluteFilePath( fileName ) << " is not a valid CubePlugin version " << PLUGIN_VERSION << endl;
                }
            }
            else
            {
                delete loader;
                Globals::debug( "PluginList" ) << "plugin " << dir.absoluteFilePath( fileName ) << " is not a valid CubePlugin version " << PLUGIN_VERSION << endl;
            }
        }
    }
    if ( cubePluginList.size() == 0 )
    {
        Globals::debug( "PluginList" ) << "no plugins found" << endl;
    }
    /*else
       {
        Globals::debugStream() << "plugins found:" << endl;
        for ( int i = 0; i < cubePluginList.size(); i++ )
        {
            Globals::debugStream() << "  " << cubePluginList.at( i ).name << endl;
        }
       }*/
}

/**
   if plugin is a valid CubePlugin, it is loaded and the loader is appended to the list pluginLoaderList
 */
bool
PluginList::loadCubePlugin( PluginData& data )
{
    CubePlugin* cubePlugin = dynamic_cast<CubePlugin*>( data.plugin );
    if ( cubePlugin )
    {
        cubePlugin->isInitialized = false;
        loadPlugin( data, cubePluginList );
    }
    return cubePlugin;
}

bool
PluginList::loadContextFreePlugin( PluginData& data )
{
    ContextFreePlugin* contextFreePlugin = dynamic_cast<ContextFreePlugin*>( data.plugin );
    if ( contextFreePlugin )
    {
        loadPlugin( data, contextFreePluginList );
    }
    return contextFreePlugin;
}


bool
PluginList::loadPlugin( PluginData& data, QList<PluginData>& list )
{
    bool ok = true;

    bool pluginIsLoaded = false;
    for ( int i = 0; i < list.size(); i++ )
    {
        if ( data.plugin ==  list[ i ].plugin )
        {
            pluginIsLoaded = true;
            break;
        }
    }

    // check, if plugin with different version already exists
    if ( !pluginIsLoaded )
    {
        int index = -1;
        for ( int i = 0; i < list.size(); i++ )
        {
            if ( data.plugin->name() == list[ i ].name )
            {
                index = i;
                break;
            }
        }

        if ( index != -1 ) // plugin already exists with different version
        {
            int version = getVersionNumber( data.plugin );

            // replace existing plugin, if version number is greater
            if ( version > getVersionNumber( list.at( index ).plugin ) )
            {
                QPluginLoader* oldLoader = list.at( index ).pluginLoader;
                oldLoader->unload();
                delete oldLoader;
                list.replace( index, data );
                Globals::debug( "PluginList" ) << "replace plugin with newer version: " << data.plugin->name() << version << data.location << endl;
            }
            else if ( version < getVersionNumber( list.at( index ).plugin ) )
            {
                Globals::debug( "PluginList" ) << "ignore older version of plugin " << data.plugin->name() << version << data.location << endl;
            }
        }
        else // insert new plugin
        {
            list.append( data );
        }
    }
    else // plugin is already loaded ( = link to same file )
    {
        data.pluginLoader->unload();
        delete data.pluginLoader;
        data.pluginLoader = 0;
        ok                = false;
    }

    return ok;
}

void
PluginList::disablePlugins()
{
    disableAlways = true;
}

/* currently unused because of bug QTBUG-39070
 * all plugins are loaded, but are activated only if an initialization function is called
   void
   PluginList::unloadPlugins()
   {
    if ( !single )
    {
        return;
    }
    else
    {
        single->unloadPluginList();
    }
   }

   void
   PluginList::unloadPluginAt( int idx )
   {
    if ( single )
    {
        QPluginLoader* loader = single->pluginLoaderList.at( idx );
        if ( loader->isLoaded() )
        {
            QString file = loader->fileName();
            loader->unload();
            delete loader;
            loader = new QPluginLoader( file );
            single->pluginLoaderList.replace( idx, loader );
            single->pluginList.replace( idx, 0 );
        }
    }
   }

   CubePlugin*
   PluginList::loadPluginAt( int idx )
   {
    CubePlugin* cubePlugin = 0;
    if ( single )
    {
        if ( single->pluginList.at( idx ) ) // already loaded
        {
            return single->pluginList.at( idx );
        }
        else // load plugin
        {
            QPluginLoader* loader = single->pluginLoaderList.at( idx );
            QObject*       plugin = loader->instance();
            if ( plugin )
            {
                cubePlugin                = qobject_cast<CubePlugin*>( plugin );
                cubePlugin->isInitialized = false;
                if ( cubePlugin )
                {
                    Globals::debugStream() << "plugin reloaded:" << cubePlugin->name() << endl;
                    single->pluginList.replace( idx, cubePlugin );
                }
                else
                {
                    Globals::debugStream() << "plugin reload failed:" << cubePlugin->name() << endl;
                }
            }
            else
            {
                Globals::debugStream() << "plugin reload failed:" << endl;
            }
        }
    }
    return cubePlugin;
   }

   void
   PluginList::unloadPluginList()
   {
    int idx = 0;
    foreach( QPluginLoader * loader, this->pluginLoaderList )
    {
        if ( loader->isLoaded() )
        {
            loader->unload();

            // cannot use loader->instance() after unload() for a second time -> segfault
            QString file = loader->fileName();
            delete loader;
            loader = new QPluginLoader( file );
            pluginLoaderList.replace( idx, loader );
            pluginList.replace( idx, 0 );
        }
        idx++;
    }
   }

 */
