#include <cassert>

#include "LaunchInfo.h"
#include "LaunchPlugin.h"
#include "CubeServices.h"

#if QT_VERSION < 0x050000
Q_EXPORT_PLUGIN2( LaunchPlugin, LaunchPlugin );
#endif

void
LaunchPlugin::version( int& major, int& minor, int& bugfix ) const
{
    major  = 1;
    minor  = 0;
    bugfix = 0;
}

QString
LaunchPlugin::getHelpText() const
{
    return "";
}

QString
LaunchPlugin::name() const
{
    return "LaunchPlugin";
}

bool
LaunchPlugin::cubeOpened( PluginServices* service )
{
    this->service = service;
    marker        = service->getTreeItemMarker( "launch" );

    connect( service, SIGNAL( treeItemIsSelected( TreeType, TreeItem* ) ),
             this, SLOT( treeItemIsSelected( TreeType, TreeItem* ) ) );
    connect( service, SIGNAL( contextMenuIsShown( TreeType, TreeItem* ) ),
             this, SLOT( contextMenuIsShown( TreeType, TreeItem* ) ) );

    LaunchInfo* launch = new LaunchInfo( service );
    if ( launch->readLaunchFile( service->getCubeBaseName() ) )
    {
        launchInfoList.append( launch );
        // todo: more launch elements in one file?
    }
    else
    {
        delete launch;
        return false;
    }

    /* create init menu items, if init items exist */
    if ( !launchInfoList.isEmpty() )
    {
        QList<LaunchInfo*> menuItems;
        for ( int i = 0; i < launchInfoList.size(); i++ )
        {
            LaunchInfo* launchInfo = launchInfoList.at( i );
            // qDebug() << "children of launchInfo "<< launchInfo->children().size();
            if ( !launchInfo->getLaunchInitMenu().isEmpty() )
            {
                menuItems.append( launchInfo );
            }
        }

        QMenu* menu = service->enablePluginMenu();
        for ( int i = 0; i < menuItems.size(); ++i )
        {
            QAction* action = menu->addAction( menuItems.at( i )->getLaunchInitMenu() );
            connect( action, SIGNAL( triggered() ), menuItems.at( i ), SLOT( launchInit() ) );
        }
    }

    return true;
}

void
LaunchPlugin::cubeClosed()
{
    foreach( LaunchInfo * info, launchInfoList )
    {
        delete info;
    }
    launchInfoList.clear();

    contextHash.clear();
}


/* create context menu items */
void
LaunchPlugin::contextMenuIsShown( TreeType type, TreeItem* item )
{
    cube::Vertex*       vertex = item->getCubeObject();
    const cube::Metric* metric = 0;
    const cube::Cnode*  cnode  = 0;

    if ( type == METRICTREE )
    {
        metric = dynamic_cast<cube::Metric* >( vertex );
    }
    else if ( type == CALLTREE )
    {
        TreeItem* metricItem = service->getSelection( METRICTREE );
        // qDebug() << "sel" << service->getSelection(METRICTREE)->name;

        metric = dynamic_cast<cube::Metric*> ( metricItem->getCubeObject() );
        cnode  = dynamic_cast<cube::Cnode*>( vertex );
    }
    else
    {
        return;
    }

    foreach( LaunchInfo * launchInfo, launchInfoList )
    {
        QStringList launchMenuEntries = launchInfo->getMenuEntries( metric, cnode );
        for ( int i = 0; i < launchMenuEntries.size(); i++ )
        {
            QAction* launchAct = service->addContextMenuItem( type, launchMenuEntries.at( i ) );

            QPair<TreeType, TreeItem*> pair = QPair<TreeType, TreeItem*> ( type, item );
            contextHash.insert( ( long )launchAct, pair );

            connect( launchAct, SIGNAL( triggered() ), this, SLOT( onLaunch() ) );

            if ( !launchInfo->isInitialized() )
            {
                launchAct->setEnabled( false );
            }
        }
    }
}

void
LaunchPlugin::onLaunch()
{
    QAction* currentMenuAction = ( QAction* )QObject::sender();
    if ( currentMenuAction == 0 )
    {
        return;
    }

    QPair<TreeType, TreeItem*> pair =  contextHash.value( ( long )currentMenuAction );

    TreeType  type = pair.first;
    TreeItem* item = pair.second;
    assert( type == METRICTREE || type == CALLTREE );

    cube::Vertex* vertex = item->getCubeObject();

    QString label = currentMenuAction->text();

    for ( int i = 0; i < launchInfoList.size(); i++ )
    {
        LaunchInfo* launchInfo = launchInfoList.at( i );

        QString command;
        if ( type == METRICTREE )
        {
            TreeItem* metricItem = item;

            cube::Metric* metric = dynamic_cast<cube::Metric* >( vertex );
            command = launchInfo->findLaunchCommand( label, metric );

            if ( !command.isEmpty() )
            {
                launchInfo->launch( command, metricItem );
            }
        }
        else if ( type == CALLTREE )
        {
            TreeItem* callItem   = item;
            TreeItem* metricItem = service->getSelection( METRICTREE );

            cube::Metric*      metric = dynamic_cast<cube::Metric*> ( metricItem->getCubeObject() );
            const cube::Cnode* cnode  = dynamic_cast<cube::Cnode*> ( vertex );

            command = launchInfo->findLaunchCommand( label, metric, cnode );

            if ( !command.isEmpty() )
            {
                launchInfo->launch( command, metricItem, callItem );
            }
        }
    }
}

/**
 * If a metric tree item is selected, mark all call tree items, which have launch information
 */
void
LaunchPlugin::treeItemIsSelected( TreeType type, TreeItem* item )
{
    ( void )item;
    if ( type == SYSTEMTREE  )
    {
        return; // selections in systemtree don't effect the marks
    }

    // clear all previously set marks
    service->clearTreeItemMarks();

    if ( service->getSelections( METRICTREE ).size() != 1 )
    {
        return; // launch is only valid, if exactly one metric item is selected
    }

    TreeItem* metricItem = service->getSelection( METRICTREE );
    for ( int i = 0; i < launchInfoList.size(); i++ )
    {
        LaunchInfo*     launchInfo = launchInfoList.at( i );
        QList<uint32_t> cnodes     = launchInfo->getCnodes();

        foreach( uint32_t cnodeId, cnodes )
        {
            TreeItem*           callItem = service->getCallTreeItem( cnodeId );
            const cube::Metric* metric   = dynamic_cast<cube::Metric*> ( metricItem->getCubeObject() );
            const cube::Cnode*  cnode    = dynamic_cast<cube::Cnode*> ( callItem->getCubeObject() );

            if ( launchInfo->existsLaunch( metric, cnode ) )
            {
                service->markTreeItem( callItem, marker );
            }
        }
    }
    service->updateTreeView( CALLTREE );
}

void
LaunchPlugin::orderHasChanged( const QList<DisplayType>& order )
{
    ( void )order;
    // todo: qDebug() << "LaunchPlugin: order changed";
}
